#include "ui.h"
#include <format>
#include <algorithm>

START_NAMESPACE_DISTRHO

static constexpr uint kInitialWidth = 800;
static constexpr uint kInitialHeight = 600;

MouseUI::MouseUI()
	: UI(kInitialWidth, kInitialHeight)
{

#ifdef DGL_NO_SHARED_RESOURCES
	createFontFromFile("sans", "/usr/share/fonts/truetype/ttf-dejavu/DejaVuSans.ttf");
#else
	loadSharedResources();
#endif
}


void MouseUI::parameterChanged(uint32_t index, float value)
{

	if (index < kParameterCount)
	{
		switch (index)
		{
		case (kHz):
			hz = value; break;
		case (kMouseX):
			x = value; break;
		case (kMouseY):
			y = value; break;
		case (kVolume):
			vol = value; break;
		}
		return;
	}

}

void MouseUI::uiIdle()
{
	repaint();
}

bool MouseUI::onMouse(const MouseEvent &ev)
{
	record = ev.press;
	repaint();
	return false; // Allow event to propagate.
}

bool MouseUI::onMotion(const MotionEvent &ev)
{
	if(!record) return false;
	
	x = std::clamp(ev.pos.getX() / (1.0 * getWidth()), 0.0, 1.0);
	y = std::clamp(ev.pos.getY() / (1.0 * getHeight()), 0.0, 1.0);
	setParameterValue(kMouseX, x);
	setParameterValue(kMouseY, y);
	repaint();

	return false;
}

void MouseUI::onResize(const ResizeEvent &ev)
{
	return UI::onResize(ev);
}

void MouseUI::onNanoDisplay()
{
	fontSize(15.0f);
	textLineHeight(1.f);

	// Numerical feedback.
	beginPath();
	fillColor(200, 200, 200);
	textBox(0.f, 15.f, 250.f,
			std::format("Frequency: {:.3f}\nNearest: {:.3f}\n",
						hz, scale.getNearestNoteNumber(hz))
				.c_str(),
			nullptr);
	closePath();

	beginPath();
	fillColor(200, 200, 200);
	textBox(0.f, 45.f, 250.f,
			std::format("x: {:.3f}\ny: {:.3f}\n",
						x * getWidth(), y * getHeight())
				.c_str(),
			nullptr);
	closePath();


	// Pen position and pressure.
	drawCircle(x, y, 0.f, 0.5f);
}

void MouseUI::drawCircle(float x, float y, float z = 0.f, float p = 1.f)
{

	static constexpr float circleRadius = 25.f;
	x *= getWidth();
	y *= getHeight();
	z = 1.f - z;

	beginPath();
	strokeColor(1.f, 1.f, 1.f, 0.5f);
	moveTo(x - z * circleRadius, y);
	lineTo(x + z * circleRadius, y);
	stroke();
	closePath();

	beginPath();
	strokeColor(1.f, 1.f, 1.f, 0.5f);
	moveTo(x, y - z * circleRadius);
	lineTo(x, y + z * circleRadius);
	stroke();
	closePath();

	beginPath();
	fillColor(1.f, 1.f, 1.f, p);
	strokeColor(255, 255, 255, 255);
	circle(x, y, circleRadius);
	fill();
	stroke();
	closePath();

	beginPath();
	strokeColor(1.f, 1.f, 1.f, z);
	circle(x, y, z * circleRadius);
	stroke();
	closePath();
}

UI *createUI()
{
	return new MouseUI();
}

END_NAMESPACE_DISTRHO
