#include <array>
#include <vector>

# define DISTRHO_DECLARE_NON_COPYABLE(ClassName) \
private:                                         \
    ClassName(ClassName&) = delete;              \
    ClassName(const ClassName&) = delete;        \
    ClassName& operator=(ClassName&) = delete;   \
    ClassName& operator=(const ClassName&) = delete;

// x that minimizes the quadratic function determined by the three points
static inline double secondOrderMinimum(int x0, float y0, int x1, float y1, int x2, float y2)
{
    // Linear system:
    // 2nd order Vandermonde matrix in x * [a b c]^T = [y0 y1 y2]^T
    // solve the 2nd order vmonde matrix, then return -b / 2a, the discriminant and xmin of qdrtic polynomial.
    // The x_i are assumed to be nonzero and distinct.
    double w0 = y0 / ( (x1 - x0) * (x2 - x0) );
    double w1 = y1 / ( (x0 - x1) * (x2 - x1) );
    double w2 = y2 / ( (x0 - x2) * (x1 - x2) );

    double a = w0 + w1 + w2;
    double b = - w0 * (x2 + x1) - w1 * (x0 + x2) - w2 * (x1 + x0);

    if ( a < 0.00001) return x1; //y not sufficiently distinct, can't get numerically meaningful answer.
    double xmin = -b / (2.0 * a);
    if ( xmin > x2 ) return x2; //We're looking for a minimum in the given interval, so just clamp it.
    if ( xmin < x0 ) return x0;
    return xmin;
};

static inline double sinc(float x)
{
    //Remove discontinuity at zero.
    if (abs( x ) < 0.001) return 1.f;
    return sinf(M_PI * x) / (M_PI * x);
};

static constexpr uint pSize = 10;
static constexpr size_t resamplerBufferSize = 1 << pSize;
static constexpr size_t resamplerMask = resamplerBufferSize - 1;

static constexpr uint dLogFactor = 3;
static constexpr uint dFactor = 1 << dLogFactor;
static constexpr size_t dBufferSize = 1 << (pSize - dLogFactor);
static constexpr size_t dBufferMask = dBufferSize - 1;

template <typename T>
class Resampler
{
private:



    // Ring buffer.
    //TODO: refactor into its own class and look up operator[] semantics.
    std::array<T, resamplerBufferSize> array = {};

    //Autocorrelations computed from scratch in a neighbourhood of some estimate.
    std::array<double, dFactor * 2> autoCorrelations;

    // Pointer for pitch correction.
    double readIdx = 0;
    // Pointer for pitch detection.
    uint writeIdx = 0;

    // Simple one-pole low pass filter for smoothing over period mismatches.
    double lpf = 0.0;

    // Calculate autocorrelation from scratch for given period
    inline double const ac(const uint per)
    {
        double ac = 0;
        uint idx = writeIdx;
        for (uint l = 0; l < per; ++l)
        {
            T x = array[idx & resamplerMask];
            T y = array[(idx - per) & resamplerMask];

            ac += (x - y) * (x - y);
            --idx;
        }
        return ac;
    }

public:

    Resampler(){};

    inline double const getAccuratePeriod(uint firstEstimate, uint secondEstimate)
    {
        //First try the two candidates for the appropriate nbhd.
        uint estimate = firstEstimate;
        double fac = ac(firstEstimate);
        if ( secondEstimate )
        {
            double altAC = ac(secondEstimate);
            if( fac > altAC )
            {
                fac = altAC;
                estimate = secondEstimate;
            }
        }

        //Get autocorrelations in the nbhd, find the minimum.
        uint minIdx = 0;
        int p = estimate - autoCorrelations.size() / 2;
        for (uint i = 0; i < autoCorrelations.size(); ++i, ++p)
        {
            autoCorrelations[i] = ac(p);
            if (autoCorrelations[i] <= fac ) {
                minIdx = i;
                estimate = p;
                fac = autoCorrelations[i];
            }
        }

        //Interpolate to get accuracy greater than one sample (yes, you need it!)
        uint x0 = estimate - 1;
        uint x1 = estimate;
        uint x2 = estimate + 1;
        double y0 = (minIdx == 0) ? ac(x0) : autoCorrelations[minIdx - 1];
        double y1 = autoCorrelations[minIdx];
        double y2 = ((minIdx + 1) >= autoCorrelations.size()) ? ac(x2) : autoCorrelations[minIdx + 1];

        return secondOrderMinimum(x0, y0, x1, y1, x2, y2);
    }

    inline void write(const T *const input, const uint32_t frames)
    {
        for (uint32_t i = 0; i < frames;
             ++i, writeIdx = (writeIdx + 1) & resamplerMask)
        {
            array[writeIdx] = input[i];
        };
    }



    //Eight tap windowed sinc filter.
    //Interpolate at readIdx - 4.0.
    inline T const read(const double rate)
    {
        const uint idx = static_cast<uint>(readIdx);
        const double frac = readIdx - idx;
        T val = 0.f;
        for( uint i = 0; i < 8; ++i )
        {
            val += sinc((idx - i - readIdx + 3.0) / rate) * array[(idx - i) & resamplerMask];
        }
        return val;
    }

    // Arguments:
    // output buffer and length
    // ratio of new sample rate to old sample rate
    // length of period in number of samples (may be a fraction)
    inline void resample(T *output, uint32_t frames, const double rate, const double period)
    {

        //Bounds of read index.
        double end = writeIdx;
        while ( readIdx > end ) { end += resamplerBufferSize; };
        const double start = end - period;

        for (uint32_t i = 0; i < frames; ++i)
        {
            output[i] = read(rate);
            readIdx += rate;

            while ( rate < 1.0 && readIdx < start )
            {
                readIdx += period;
            }

            while ( readIdx > end )
            {
                readIdx -= period;
            }

        };

        //Prevent read index overflow / precision loss.
        while (readIdx > resamplerBufferSize) {readIdx -= resamplerBufferSize;};
    }

    DISTRHO_DECLARE_NON_COPYABLE(Resampler)
};

template <typename T>
class Detector
{
    
    std::array<T, dBufferSize> array = {};
    // Pointer for pitch detection.
    uint writeIdx = 0; 
    // Detectable periods.
    std::array<double, dBufferSize / 2> squares = {};
    std::array<double, dBufferSize / 2> crosses = {};
    static constexpr uint minPeriod = 16;
    static constexpr uint maxPeriod = dBufferSize / 2;

public:

    Detector(){};

    // Read inputFrames from input buffer, downsample by factor and write to our ring buffer
    //  TODO: implement downsampling filter. This will probably have way too much antialiasing!
    inline void downsample(const T* const input, const uint32_t inputFrames)
    {
        for (uint32_t i = 0; i < inputFrames;
             i += dFactor, writeIdx = (writeIdx + 1) & dBufferMask)
        {
            uint32_t j = (i > 3) ? i - 4 : 0;
            array[writeIdx] = 0.5f * (input[i] + input[j]); //Lousy lowpass filter just for now.  
        };
    }

    std::array<uint, 2> detectPitch(uint32_t frames)
    {
        frames /= dFactor;
        uint idx = writeIdx - frames;
        std::array<uint, 2> pitches{};

        //Update autocorrelations.
        while (frames)
        {
            for (uint per = 3; per < squares.size(); ++per)
            {
                T x = array[idx & dBufferMask];
                T y = array[(idx - per) & dBufferMask];
                T z = array[(idx - 2 * per) & dBufferMask];
                //Autocorrelation and RMS computation.
                squares[per] += x * x - z * z;
                crosses[per] += x * y - y * z;
            }
            --frames;
            ++idx;
        }

        //Search for local minima of autocorrelation function.
        for ( uint per = 3; per < squares.size() - 1; ++per)
        {
            if ( 0.49 * squares[per] < crosses[per] )
            {
                double a = 2.0 * crosses[per - 1] - squares[per - 1];
                double b = 2.0 * crosses[per] - squares[per];
                double c = 2.0 * crosses[per + 1] - squares[per + 1];
                if ( (b > a) && (b > c) )
                {
                    if (pitches[0])
                    {
                        pitches[1] = dFactor * per;
                        return pitches;
                    }
                    else
                    {
                        pitches[0] = dFactor * per;
                        if (per * 2 > squares.size() ) return pitches;
                    }
                }
            }
        }

        return pitches;

    }

     DISTRHO_DECLARE_NON_COPYABLE(Detector)
};