#include "filter.h"

void Resonator::clear()
{
  xp = xpp = yp = ypp = 0;
}

inline float Resonator::process(float x)
{
  float y = cx * x + cxp * xp - cypp * ypp - cyp * yp;
  xpp = xp;
  xp = x;
  ypp = yp;
  yp = y;
  return y;
}

void Resonator::set(double _cx, double _cxp, double _cyp, double _cypp)
{
  cx = _cx;
  cxp = _cxp;
  cyp = _cyp;
  cypp = _cypp;
}

static constexpr double sampleRate = 48000.0;

template <uint N>
void ParallelResonators<N>::process(float **outputs, const float **inputs, uint32_t frames)
{
  for (uint chn = 0; chn < 2; ++chn)
  {
    for (auto &r : res[chn])
    {
      for (uint32_t j = 0; j < frames; ++j)
      {
        outputs[chn][j] += r.process(inputs[chn][j]);
      }
    }
  }
};

template <uint N>
void set(std::array<double, N> &angle, std::array<double, N> &radius)
{

  //Get poles.
  for ( uint i = 0; i < N; ++i )
  {
    radius[i] = exp(-M_PI * radius[i] / sampleRate);
    angle[i] = M_PI * 2.0 * angle[i] / sampleRate ;
  }

  //Partial fraction expansion.
  //Compute residue of each pole, then combine with complex conjugate
  //to form a real second-order section.
  //Pass this to the associated filter.
  for ( uint i = 0; i < N; ++i )
  {

    //Residue
    double x = 0.0;
    double y = 0.0;
    for (uint j = 0; j < N; ++j)
    {
      if ( i == j ) continue;

      // 1 - p_j / p_i
      double re = 1.0 - radius[j] * cos(angle[j] - angle[i]) / radius[i];
      double im = 0.0 - radius[j] * sin(angle[j] - angle[i]) / radius[i];

      // (x + iy) * (re + i im) = re * x - y * im
      x = re * x - y * im;
      y = im * x + y * re;

    }

    //Pole
    double u = radius[i] * cos(angle[i]);
    double v = radius[i] * sin(angle[i]);

    //Multiply complex conjugate pairs to get real biquad coefficients.
    double cxp = -2.0 * (u * x + v * y);
    double cx = 2.0 * x;
    double cypp = radius[i] * radius[i];
    double cyp = -2.0 * u;

    //Save biquad coefs.
    res[0][i].set(cx, cxp, cyp, cypp);
    res[1][i].set(cx, cxp, cyp, cypp);
  }

};